;;; freecell.scm -- Free Cell game for AisleRiot.

;; Copyright (C) 1998 Changwoo Ryu
;; Copyright (c) 1999, 2000 Matthew Wilcox <matthew@wil.cx>

;; Author: Changwoo Ryu <cwryu@adam.kaist.ac.kr>

;; This program is free software; you can redistribute it and'or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

; The original version of this uses non-standard names for the slots.
; There may still be some remnants of this, so the translation is:
; freecells = reserve
; fields = tableau

;;
;; Game Options
;;
(define option-auto-move #t)
(define option-one-by-one #f)

;;
;; Constants
;;
(define fields '(8 9 10 11 12 13 14 15))

;;
;; Utilities
;;

(define (freecell? slot)
	(member slot reserve)
)

(define (field? slot)
	(member slot fields)
)

(define (count-empty-slots list)
	(define (f slot)
		(if (empty-slot? slot)
			1
			0
		)
	)
	(apply + (map f list))
)

;;
;; Utilities for freecells
;;

;; The total number of empty freecells
(define (empty-freecell-number)
	(count-empty-slots reserve)
)

;; An empty freecell's id, if any
(define (any-empty-freecell)
	(find-empty-slot reserve)
)

;;
;; Utilities for fields
;;

(define (to-tableau? lower upper)
	(and (not (color-eq? lower upper)) (card-follows? lower upper))
)

(define (field-sequence? card-list)
	(or
		(> 1 (length card-list))
		(and
			(to-tableau? (car card-list) (cadr card-list))
			(field-sequence? (cdr card-list))
		)
	)
)

(define (empty-field-number start end)
	(count-empty-slots (delq start (delq end fields)))
)

;;
;; How to move cards
;;

(define (move-to-foundation card-list)
	(and
		(= (length card-list) 1)
		(move-card-to-foundation (car card-list))
	)
)

(define (move-card-to-foundation card)
	(define foundation-id (find-foundation-slot card))
	(and
		foundation-id
		(add-to-score! 1)
		(add-card! foundation-id card)
	)
)

(define (max-cards-move start-slot end-slot)
	(*
		(+ (empty-freecell-number) 1)
		(integer-expt 2 (empty-field-number start-slot end-slot))
	)
)

(define (move-to-field card-list start field-id)
	(and
		(<= (length card-list) (max-cards-move start field-id))
		(or
			(empty-slot? field-id)
			(next-tableau-card? (car (reverse card-list)) field-id)
		)
		(add-cards! field-id card-list)
	)
)

(define (move-to-freecell card-list freecell-id)
	(and
		(= (length card-list) 1)
		(move-card-to-freecell (car card-list) freecell-id)
	)
)

(define (move-card-to-freecell card freecell-id)
	(and
		freecell-id
		(empty-slot? freecell-id)
		(add-card! freecell-id card)
	)
)

;;
;; Callbacks & Initialize the game
;;

;; Set up a new game.
(define (new-game)
	(initialize-playing-area)
	(set-ace-low)
	(make-standard-deck)
	(shuffle-deck)

	(add-slots)

	(set! reserve '(0 1 2 3))
	(set! foundations '(4 5 6 7))
	(set! tableau '(8 9 10 11 12 13 14 15))

	(deal-cards-face-up-from-deck DECK (append tableau tableau
		tableau tableau tableau tableau '(8 9 10 11)))

	(read-foundations)
	(list 8 3)
)

(define (add-slots)
	(add-normal-slot '())
	(add-normal-slot '())
	(add-normal-slot '())
	(add-normal-slot '())

	(set! HORIZPOS (+ 21 HORIZPOS))

	(add-normal-slot '())
	(add-normal-slot '())
	(add-normal-slot '())
	(add-normal-slot '())

	(add-carriage-return-slot)

	(do
		((i 0 (+ 1 i)))
		((= 8 i))
		(add-extended-slot '() down)
		(set! HORIZPOS (+ 3 HORIZPOS))
	)
)

(define (button-pressed slot card-list)
	(or
		(freecell? slot)
		(and
			(field? slot)
			(field-sequence? card-list)
		)
	)
)

(define (button-released start-slot card-list end-slot)
	(and
		(not (= start-slot end-slot))
		(cond
			((foundation? end-slot) (move-to-foundation card-list))
			((field? end-slot)
			 (move-to-field card-list start-slot end-slot))
			((freecell? end-slot)
			 (move-to-freecell card-list end-slot))
		)
		(auto-move-cards (append reserve fields))
	)
)
  
(define (button-clicked slot) #f)

(define (button-double-clicked slot)
	(and
		(not (empty-slot? slot))
		(move-card-to-freecell (get-top-card slot) (any-empty-freecell))
		(remove-card slot)
		(auto-move-cards (append reserve fields))
	)
)

;; Auto move stuff

(define min-red 0)
(define min-black 0)

(define (get-foundation-value suit)
	(define slot (find-slot-by-suit suit foundations))
	(if slot
		(get-value (get-top-card slot))
		0
	)
)

(define (read-foundations)
	(set! min-black (min
		(get-foundation-value club) (get-foundation-value spade)))
	(set! min-red (min
		(get-foundation-value heart) (get-foundation-value diamond)))
)

(define (auto-move-cards slot-list)
	(define (auto-move? card)
		(<= (get-value card)
			(if (is-red? card)
				(min (+ 2 min-black) (+ 3 min-red))
				(min (+ 2 min-red) (+ 3 min-black))
			)
		)
	)

	(define (f slot)
		(and
			(not (empty-slot? slot))
			(let* ( (card (get-top-card slot))
				(end-slot (find-foundation-slot card)) )
				(and
					end-slot
					(auto-move? card)
					(remove-card slot)
					(move-card-to-foundation card)
				)
			)
		)
	)

	(read-foundations)
	(and
		(or-map f slot-list)
		(auto-move-cards slot-list)
	)
	#t
)

;; Condition for fail -- no more cards to move
(define (game-over)
  ; (FIXME)
  (not (game-won)))

;; Condition for win -- all the cards in foundations
(define (game-won)
	(foundations-full?)
)

(define (get-hint)
  ; (FIXME)
  #f)
 
(define (get-options) 
  #f)
;  '(("Auto move to foundation" option-auto-move)
;    ("Move one by one" option-one-by-one)))

(define (apply-options options) 
  #f)
;  (set! option-auto-move (cadar options))
;  (set! option-auto-move (cadadr options)))

(define (timeout) 
  ; (FIXME)
  #f)

(set-lambda new-game button-pressed button-released button-clicked button-double-clicked game-over game-won get-hint get-options apply-options timeout)
