; AisleRiot - sol.scm
; Copyright (C) 1998 Jonathan Blandford <jrb@mit.edu>
;
; This game is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
; USA

;; Constants:

(define jack 11)
(define queen 12)
(define king 13)
(define ace 1)
(define joker 0)

(define club 0)
(define diamond 1)
(define heart 2)
(define spade 3)

(define black 0)
(define red 1)

(define down 0)
(define right 1)

;; Global variables:

(define DECK '())

(define foundations '())
(define tableau '())
(define reserve '())
(define stock '())
(define waste '())
(define moveable '())

;; NEW-GAME PROCEDURES
; these may be used in game files during the new-game procedure.

; This procedure MUST be called at the start of the new-game procedure.
(define (initialize-playing-area)
  (reset-surface)
  (set! FLIP-COUNT 0)
  (set! DEAL-COUNT 1)
  (set! SLOTS 0)
  (set! HORIZPOS (get-horiz-start))
  (set! VERTPOS (get-vert-start))
  (set! IN-GAME #f)
  (set! MOVE '())
  (set-statusbar-message "")
  (set! HISTORY '()))

(define (set-deck! card-list)
	(set! DECK card-list)
)

; create a 52 card deck (puts list of cards into DECK)
(define (make-standard-deck)
	(if (= ace 14)
		(make-restricted-deck 2)
		(make-restricted-deck ace)
	)
)

; create a card deck from first to ace or king, depending which is high.
(define (make-restricted-deck first)
	(set-deck! (if (= ace 14)
		(make-deck-list first ace)
		(make-deck-list first king)
	))
)

; create a 54 card deck with 2 jokers.
(define (make-joker-deck)
	(set-deck! (cons (make-card joker club) (cons (make-card joker diamond)
		(if (= ace 14)
			(make-deck-list 2 ace)
			(make-deck-list ace king)
		)
	)))
)

; create a double deck of 104 cards (puts list of cards into DECK)
(define (make-standard-double-deck)
	(if (= ace 14)
		(set! DECK (append (make-deck-list 2 ace) (make-deck-list 2 ace)))
		(set! DECK (append (make-deck-list ace king) (make-deck-list ace king)))
	)
)

(define (make-restricted-double-deck first)
	(if (= ace 14)
		(set! DECK (append (make-deck-list first ace) (make-deck-list first ace)))
		(set! DECK (append (make-deck-list first king) (make-deck-list first king)))
	)
)

; makes a deck from first to last
(define (make-deck-list first last)
	(define (f value suit deck)
		(define newdeck (cons (make-card value suit) deck))
		(if (= first value)
			(if (= club suit)
				newdeck
				(f last (1- suit) newdeck)
			)
			(f (1- value) suit newdeck)
		)
	)
	(f last spade '())
)

; shuffle the card list in DECK
(define (shuffle-deck)
	(define vec (list->vector DECK))
	(define (f src len)
		(let* (
			(dest (+ src (random len)))
			(card (vector-ref vec dest))
		)
		(vector-set! vec dest (vector-ref vec src))
		(vector-set! vec src card)
		(if (< 2 len)
			(f (1+ src) (1- len))
			vec
		)
		)
	)
	(f 0 (vector-length vec))
	(set! DECK (vector->list vec))
)

; The playing area is divided into slots, where cards can be placed.
; Each slot can hold any amount of cards.  The slots are identified 
; using numbers assigned in order of their creation. The deck of cards
; held in DECK should be assigned to one of the slots on creation.
; (You may then create another deck and place it in a later slot).
;
; The slots are added to the board from left to right until the
; add-carriage-return-slot procedure is called, which starts a new line.
; A space may be added using the add-blank-slot procedure. These false
; slots are not assigned identifiers.

(define (add-blank-slot)
  (get-and-increment-position))

(define (add-carriage-return-slot)
  (linefeed-position))

; The real slots come in three varieties:
; A slot in which only the topmost card is visible:
(define (add-normal-slot cards)
  (add-slot (set-tag! (new-slot cards 
				(list 'normal (get-and-increment-position))))))

; A slot in which all the cards are visible, arranged as an overlapped pile:
; (either proceeding to the right or down).
(define (add-extended-slot cards direction)
  (if (= right direction)
      (add-slot (set-tag! (new-slot cards 
				    (list 'expanded-right 
					  (get-and-increment-position)))))
      (add-slot (set-tag! (new-slot cards 
				    (list 'expanded 
					  (get-and-increment-position)))))))

; A slot in only the n topmost cards are visible:
(define (add-partially-extended-slot cards direction n)
  (if (= right direction)
      (add-slot (set-tag! (new-slot cards 
				    (list 'partially-expanded-right 
					  (get-and-increment-position) n))))
      (add-slot (set-tag! (new-slot cards 
				    (list 'partially-expanded 
					  (get-and-increment-position) n))))))

; Cards may be dealt off one slot (usually the one containing the deck)
; and onto a list of other slots using these procedures:
(define (deal-cards target-slot-id slot-list)
	(if (null? slot-list)
		#t
		(begin
			(add-card! (car slot-list) (remove-card target-slot-id))
			(deal-cards target-slot-id (cdr slot-list))
		)
	)
)

(define (deal-cards-face-up target-slot-id slot-list)
	(if (null? slot-list)
		#t
		(begin
			(add-card! (car slot-list)
				(make-visible (remove-card target-slot-id)))
			(deal-cards-face-up target-slot-id (cdr slot-list))
		)
	)
)

(define (deal-cards-face-up-from-deck deck slot-list)
	(if (null? slot-list)
		#t
		(begin
			(add-card! (car slot-list) (make-visible (car deck)))
			(deal-cards-face-up-from-deck (cdr deck) (cdr slot-list))
		)
	)
)

;; GENERAL GAME PROCEDURES
; these may be used in game files at any time.

;; Procedures that change slot contents:

; turn the top card of a slot over (face up to face down and vice versa)
(define (flip-top-card slot-id)
  (add-card! slot-id (flip-card (remove-card slot-id))))

; turn the top card of a slot face side up 
(define (make-visible-top-card slot-id)
  (add-card! slot-id (make-visible (remove-card slot-id))))

; add a card onto the top of a slot
(define (add-card! slot-id card)
  (set-cards! slot-id (cons card (get-cards slot-id))))

; add a list of cards onto the top of a slot
(define (add-cards! slot-id cards)
  (set-cards! slot-id (append cards (get-cards slot-id))))

; remove (and return) the top card from a slot
(define (remove-card slot-id)
  (let ((cards (get-cards slot-id)))
    (set-cards! slot-id (cdr cards))
    (car cards)))

;; Utilities

; the number of redeals allowed.
(define (set-redeals n)
	(set! FLIP-COUNT n)
)

; the number of cards dealt at a time.
(define (set-deal-count n)
	(set! DEAL-COUNT n)
)

; deal a card from the stock-slot to the waste-slot.
; when the stock slot is empty than the waste slot will be flipped back
; onto the stock unless the flip limit has been reached.
(define (flip-stock)
	(if (empty-slot? stock)
		(and
			(< 0 FLIP-COUNT)
			(set! FLIP-COUNT (1- FLIP-COUNT))
			(flip-deck waste)
		)
		(or (let loop ((i DEAL-COUNT))
			(and
				(> i 0)
				(not (empty-slot? stock))
				(add-card! waste (flip-card (remove-card stock)))
				(loop (1- i))
			)
		    )
			#t
		)
	)
)

; turn the cards in the waste slot over and add them to the stock-slot.
(define (flip-deck waste-slot)
  (and (not (empty-slot? waste-slot))
       (add-card! stock (flip-card (remove-card waste-slot)))
       (or (flip-deck waste-slot)
	   #t)))

;; Procedures for manipulating cards:

; NB: In order to use these procedures you must remove the cards 
;     from their slots and then replace them after applying the procedure 
;     (as in the make-top-card-visible procedure above)
(define (flip-card card)
  (list (car card) (cadr card) (not (caddr card))))

(define (make-visible card)
  (list (car card) (cadr card) #t))

;; Procedures that provide information only:

; card procedures
(define (is-visible? card)
  (caddr card))

(define (get-suit card) 
  (cadr card))

(define (get-color card)
  (cond ((eq? (get-suit card) club) black)
	((eq? (get-suit card) spade) black)
	((eq? (get-suit card) heart) red)
	((eq? (get-suit card) diamond) red)
	(#t "Unknown color")))

(define (get-value card) 
  (car card))

; slot procedures
(define (get-cards slot-id)
  (cadr (get-slot slot-id)))

(define (empty-slot? slot-id)
  (null? (get-cards slot-id)))

(define (get-top-card slot-id)
  (let ((cards (get-cards slot-id)))
    (if (null? cards)
	cards
	(car cards))))

(define (get-bottom-card slot-id)
	(let ((cards (get-cards slot-id)))
	(if (null? cards)
		cards
		(car (reverse cards))
	))
)

;; Utilities - need more of these:

(define (suit-eq? card1 card2)
	(= (get-suit card1) (get-suit card2))
)

(define (color-eq? card1 card2)
	(eq? (get-color card1) (get-color card2))
)

(define (value-eq? card1 card2)
	(eq? (get-value card1) (get-value card2))
)
	

(define (cards-eq? card1 card2)
	(and
		(suit-eq? card1 card2)
		(value-eq? card1 card2)
	)
)

(define (is-red? card)
  (eq? red (get-color card)))

(define (is-black? card)
  (eq? black (get-color card)))

(define (is-joker? card)
  (= (get-value card) joker))

(define (set-ace-low)
	(set! ace 1)
	(set! BASE ace)
	(set! WRAP #f)
)

(define (set-ace-high)
	(set! ace 14)
	(set! BASE 2)
	(set! WRAP #t)
)

(define (set-base-card! value)
	(set! BASE value)
	(set! WRAP #t)
)

(define (set-wrap)
	(set! WRAP #t)
)

(define (foundation? slot)
	(member slot foundations)
)

; test to see if foundations are full.  A good game-won test.
(define (foundations-full?)
	(define (foundation-full? slot)
		(= 13 (length (get-cards slot)))
	)
	(and-map foundation-full? foundations)
)

; #t if the second card has value 1 greater than the first, taking into
; account wrap-around (so ace follows king).
(define (card-follows? first second)
	(if WRAP
		(= (remainder (1+ (get-value first)) 13)
			(remainder (get-value second) 13)
		)
		(= (1+ (get-value first)) (get-value second))
	)
)

; #t if first and second are in the same suit and second follows first.
(define (suit-next? first second)
	(and (suit-eq? first second) (card-follows? first second))
)

; test to see whether a card can be placed on a foundation slot
(define (next-foundation-card? card slot)
	(define top (get-top-card slot))
	(if (null? top)
		(= BASE (get-value card))
		(suit-next? top card)
	)
)

; test to see whether a card can be placed on a tableau slot
(define (next-tableau-card? card slot)
	(define top (get-top-card slot))
	(if (null? top)
		(= king (get-value card))
		(and
			(not (color-eq? card top))
			(card-follows? card top)
		)
	)
)

(define (find-empty-slot slots)
	(or-map (lambda (s) (and (empty-slot? s) s)) slots)
)

; slot number which holds the suit, or #f if no slot does.
(define (find-slot-by-suit suit slots)
	(define (s slot)
		(and	(not (empty-slot? slot))
			(= suit (get-suit (get-top-card slot)))
			slot
		)
	)
	(or-map s slots)
)

; returns a foundation slot which can hold the card
; or #f if no such slot exists.
(define (find-foundation-slot card)
	(define (f slot)
		(and
			(next-foundation-card? card slot)
			slot
		)
	)

	(or-map f foundations)
)

(define (check-same-suit-list card-list)
  (or (< (length card-list) 2)
      (and (suit-eq? (car card-list) (cadr card-list))
	   (check-same-suit-list (cdr card-list)))))

(define (check-straight-descending-list card-list)
  (or (< (length card-list) 2)
      (and (card-follows? (car card-list) (cadr card-list))
	   (check-straight-descending-list (cdr card-list)))))

; debugging aid:
(define (display-list . objs)
  (map display objs) (newline))

; hint procedures: these will be superceeded shortly.
(define (get-value-name value)
  (cond ((eq? value ace) "ace")
	((eq? value 2) "two")
	((eq? value 3) "three")
	((eq? value 4) "four")
	((eq? value 5) "five")
	((eq? value 6) "six")
	((eq? value 7) "seven")
	((eq? value 8) "eight")
	((eq? value 9) "nine")
	((eq? value 10) "ten")
	((eq? value jack) "jack")
	((eq? value queen) "queen")
	((eq? value king) "king")
	(#t "Unknown value")))

(define (get-suit-name suit)
  (cond ((eq? suit club) "clubs")
	((eq? suit spade) "spades")
	((eq? suit heart) "hearts")
	((eq? suit diamond) "diamonds")
	(#t "Unknown suit")))

(define (get-joker-name card) 
  (if (is-black? card) "black joker" "red joker"))

(define (get-name card)
  (if (is-joker? card)
      (get-joker-name card)
      (string-append (get-value-name (get-value card)) 
		     " of "
		     (get-suit-name (get-suit card)))))

; return a hint if the top card of slot-id can be placed on a foundation
; pile.  Otherwise returns #f.  Use as (or-map check-to-foundation slot-list)
(define (check-to-foundation slot-id)
	(define card (get-top-card slot-id))
	(and
		(not (null? card))
		(is-visible? card)
		(let ((dest (find-foundation-slot card)))
			(and
				dest
				(if (empty-slot? dest)
					(list 3 (get-name card)
						"to an empty foundation")
					(list 1 (get-name card)
						(get-name (get-top-card dest)))
				)
			)
		)
	)
)

; find a king, and return a hint
(define (here-kingy-kingy slot-id)
	(define card-list (get-cards slot-id))
	(define (f card)
		(and
			(is-visible? card)
			(= king (get-value card))
			(list 2 (get-name card) "an empty slot")
		)
	)
	(and
		(> (length card-list) 1)
		; skip the bottom card, and reverse order finds better cards
		(or-map f (cdr (reverse card-list)))
	)
)

(define (deal-hint)
	(if (empty-slot? stock)
		(and
			(not (empty-slot? waste))
			(< 0 FLIP-COUNT)
			(list 0 "Move waste back to stock")
		)
		(if (= 1 DEAL-COUNT)
			(list 0 "Deal a card from the deck")
			(list 0 "Deal more cards from the deck")
		)
	)
)

(define (get-base-string)
	(string-append "Base Card:  "
		(cond
			( (= BASE 1)  "Ace" )
			( (= BASE 11) "Jack" )
			( (= BASE 12) "Queen" )
			( (= BASE 13) "King" )
			( (< BASE 11) (number->string BASE) )
			( else "Unknown" )
		)
	)
)

(define (get-stock-no-string)
	(string-append "Stock left:  "
		(number->string (length (get-cards stock))))
)

(define (get-reserve-no-string)
	(string-append "Reserve left:  "
		(number->string (length (get-cards reserve))))
)

(define (get-redeals-string)
	(string-append "Redeals left:  " (number->string FLIP-COUNT))
)

(define (move-n-cards! start-slot end-slot cards)
  (add-cards! end-slot cards))

(define (remove-n-cards slot-id n)
  (set-cards! slot-id (nthcdr n (get-cards slot-id))))


(define (set-cards! slot-id new_cards)
;  (if IN-GAME 
;      (record-move slot-id new_cards (get-cards slot-id)))
  (set-cards-c! slot-id new_cards))

(define (make-card value suit)
  (list value suit #f))

(define (make-visible-card value suit)
  (list value suit #t))

(define (new-slot deck placement)
  (list #f deck placement))

(define (set-tag! slot)
  (set! SLOTS (1+ SLOTS))
  (cons (1- SLOTS) (cdr slot)))

(define (get-and-increment-position)
  (let ((retval (list HORIZPOS VERTPOS)))
    (set! HORIZPOS (+ HORIZPOS (get-horiz-offset)))
    retval))

(define (linefeed-position)
  (set! HORIZPOS (get-horiz-start))
  (set! VERTPOS (+ VERTPOS (get-vert-offset))))

(define (register-undo-function function data)
  (set! MOVE (cons '(function data) (cdr MOVE))))

; common lisp procedure not provided in guile 1.3
(define (nthcdr n lst)
  (if (zero? n) lst (nthcdr (1- n) (cdr lst))))

;; INTERNAL procedures

; global variables
(define FLIP-COUNT 0)
(define SLOTS 0)
(define HORIZPOS 0)
(define VERTPOS 0)
(define MOVE '())
(define HISTORY '())
(define FUTURE '())
(define IN-GAME #f)
(define BASE ace)
(define WRAP #t)
(define DEAL-COUNT 1)

; called from C:
(define (start-game)
  (set! IN-GAME #t))

; called from C:
(define (end-move)
  (if (not (= 0 (length MOVE)))
      (begin
	(set! HISTORY (cons MOVE HISTORY))
	(set! FUTURE '())
	(set! MOVE '()))))

(define (return-cards card-positions slot-id)
  (and (not (= 0 (length card-positions)))
       (set-cards! slot-id (car card-positions))
       (return-cards (cdr card-positions) (1+ slot-id))))

(define (give-status-message)
  #t)

(define (eval-move move)
  (return-cards (caddr move) 0)
  ((car move) (cadr move))
  (give-status-message))

; called from C:
(define (undo)
  (and (not (null? HISTORY))
       (record-move -1 '())
       (eval-move (car HISTORY))
       (set! FUTURE (cons MOVE FUTURE))
       (set! HISTORY (cdr HISTORY))
       (set! MOVE '())))

; called from C:
(define (redo)
  (and (not (null? FUTURE))
       (record-move -1 '())
       (eval-move (car FUTURE))
       (set! HISTORY (cons MOVE HISTORY))
       (set! FUTURE (cdr FUTURE))
       (set! MOVE '())))

(define (undo-func data)
  (set-score! (car data))
  (set! FLIP-COUNT (cadr data)))
;(register-undo-function undo-func '(score FLIP-COUNT))
	     
(define (snapshot-board slot-id moving-slot old-cards)
  (cond ((>= slot-id SLOTS)
	 '())
	((= slot-id moving-slot)
	 (cons old-cards 
	       (snapshot-board (1+ slot-id) moving-slot old-cards)))
	(#t
	 (cons (get-cards slot-id) 
	       (snapshot-board (1+ slot-id) moving-slot old-cards)))))

; called from C:
(define (record-move slot-id old-cards)
  (set! MOVE (list undo-func 
		   (list (get-score) FLIP-COUNT)
		   (snapshot-board 0 slot-id old-cards))))

; called from C:
(define (discard-move)
  (set! MOVE '()))
